﻿using System.Collections.Generic;
using System.Text;
using UnityEngine;
using UnityEngine.UI;
using uPLibrary.Networking.M2Mqtt.Messages;

/// <summary>
/// https://mntolia.com/mqtt-qos-levels-explained/#:~:text=MQTT%20QoS%20Level%200%20%E2%80%93%20At%20Most%20Once,broker%20receives%20the%20message%20sent%20by%20the%20client.
/// Examples for the M2MQTT library (https://github.com/eclipse/paho.mqtt.m2mqtt),
/// </summary>
namespace M2MqttUnity.Examples
{
    /// <summary>
    /// Script for testing M2MQTT with a Unity UI
    /// </summary>
    public class M2MqttUnityTest : M2MqttUnityClient
    {
        [Tooltip("Set this to true to perform a testing cycle automatically on startup")]
        public bool autoTest = false;
        [Header("用户界面")]
        public InputField messageInputField;
        public InputField consoleInputField;
        public Toggle encryptedToggle;
        public InputField addressInputField;
        public InputField portInputField;
        public Button connectButton;
        public Button disconnectButton;
        public Button testPublishButton;
        public Button clearButton;
        public Toggle[] topicToggles;//用来存放Toggle
        public string[] subTopics = new string[2];
        private string topic;

        private List<string> eventMessages = new List<string>();
        private bool updateUI = false;

        /// <summary>
        /// 测试发布
        /// </summary>
        public void TestPublish()
        {
            if(string.IsNullOrEmpty(topic)) topic = "client/manage";
            client.Publish(topic, Encoding.UTF8.GetBytes(messageInputField.text), MqttMsgBase.QOS_LEVEL_EXACTLY_ONCE, false);

            Debug.Log($"{topic}发布了一条消息:{messageInputField.text}");
            AddUIMessage("发布消息："+messageInputField.text);
        }

        /// <summary>
        /// 设置IP地址或URL
        /// </summary>
        /// <param name="brokerAddress">输入框字符串</param>
        public void SetBrokerAddress(string brokerAddress)
        {
            if (addressInputField && !updateUI)
            {
                this.brokerAddress = brokerAddress;
            }
        }

        /// <summary>
        /// 设置端口
        /// </summary>
        /// <param name="brokerPort">输入框字符串</param>
        public void SetBrokerPort(string brokerPort)
        {
            if (portInputField && !updateUI)
            {
                int.TryParse(brokerPort, out this.brokerPort);
            }
        }

        /// <summary>
        /// 设置是否加密
        /// </summary>
        /// <param name="isEncrypted">Toogle isOn</param>
        public void SetEncrypted(bool isEncrypted)
        {
            this.isEncrypted = isEncrypted;
        }

        /// <summary>
        /// 设置UI显示消息
        /// </summary>
        /// <param name="msg"></param>
        public void SetUIMessage(string msg)
        {
            if (consoleInputField != null)
            {
                //  consoleInputField.text = msg;
                updateUI = true;
            }
        }
        /// <summary>
        /// UI显示增加消息
        /// </summary>
        /// <param name="msg"></param>
        public void AddUIMessage(string msg)
        {
            if (consoleInputField != null)
            {
                consoleInputField.text += msg + "\n";
                updateUI = true;
            }
        }

        protected override void OnConnecting()
        {
            base.OnConnecting();
            SetUIMessage("Connecting to broker on " + brokerAddress + ":" + brokerPort.ToString() + "...\n");
        }

        protected override void OnConnected()
        {
            base.OnConnected();
            SetUIMessage("Connected to broker on " + brokerAddress + "\n");

            if (autoTest)
            {
                TestPublish();
            }
        }

        protected override void SubscribeTopics()
        {
            client.Subscribe(subTopics, new byte[]
            {
                MqttMsgBase.QOS_LEVEL_EXACTLY_ONCE ,
                MqttMsgBase.QOS_LEVEL_EXACTLY_ONCE
            });
        }

        protected override void UnsubscribeTopics()
        {
            client.Unsubscribe(subTopics);
        }

        protected override void OnConnectionFailed(string errorMessage)
        {
            AddUIMessage("CONNECTION FAILED! " + errorMessage);
        }

        protected override void OnDisconnected()
        {
            AddUIMessage("Disconnected.");
        }

        protected override void OnConnectionLost()
        {
            AddUIMessage("CONNECTION LOST!");
        }

        private void UpdateUI()
        {
            if (client == null)
            {
                if (connectButton != null)
                {
                    connectButton.interactable = true;
                    disconnectButton.interactable = false;
                    testPublishButton.interactable = false;
                }
            }
            else
            {
                if (testPublishButton != null)
                {
                    testPublishButton.interactable = client.IsConnected;
                }
                if (disconnectButton != null)
                {
                    disconnectButton.interactable = client.IsConnected;
                }
                if (connectButton != null)
                {
                    connectButton.interactable = !client.IsConnected;
                }
            }
            if (addressInputField != null && connectButton != null)
            {
                addressInputField.interactable = connectButton.interactable;
                addressInputField.text = brokerAddress;
            }
            if (portInputField != null && connectButton != null)
            {
                portInputField.interactable = connectButton.interactable;
                portInputField.text = brokerPort.ToString();
            }
            if (encryptedToggle != null && connectButton != null)
            {
                encryptedToggle.interactable = connectButton.interactable;
                encryptedToggle.isOn = isEncrypted;
            }
            if (clearButton != null && connectButton != null)
            {
                clearButton.interactable = connectButton.interactable;
            }
            updateUI = false;
        }

        protected override void Start()
        {
            SetUIMessage("Ready.");
            updateUI = true;
            base.Start();

            for (int i = 0; i < topicToggles.Length; i++)
            {
                Toggle toggle = topicToggles[i];//循环遍历添加
                toggle.onValueChanged.AddListener((bool value) => OnValueChange(toggle));
            }
        }

        protected override void DecodeMessage(string topic, byte[] message)
        {
            string msg = Encoding.UTF8.GetString(message);
            Debug.Log("收到消息: " + msg);
            StoreMessage(msg);
            if (topic == "M2MQTT_Unity/test")
            {
                if (autoTest)
                {
                    autoTest = false;
                    Disconnect();
                }
            }
        }

        /// <summary>
        /// 存储消息
        /// </summary>
        private void StoreMessage(string eventMsg)
        {
            eventMessages.Add(eventMsg);
        }
        /// <summary>
        /// 处理消息显示到UI上
        /// </summary>
        private void ProcessMessage(string msg)
        {
            AddUIMessage("收到消息: " + msg);
        }

        protected override void Update()
        {
            base.Update(); // call ProcessMqttEvents()
            if (eventMessages.Count > 0)
            {
                foreach (string msg in eventMessages)
                {
                    ProcessMessage(msg);
                }
                eventMessages.Clear();
            }
            if (updateUI)
            {
                UpdateUI();
            }
        }

        private void OnDestroy()
        {
            Disconnect();
        }

        private void OnValidate()
        {
            if (autoTest)
            {
                autoConnect = true;
            }
        }

        void OnValueChange(Toggle t)
        {
            if (t.isOn)
            {
                switch (t.name)
                {
                    case "manage":
                        topic = "client/manage";
                        subTopics[0] = "client/manage";
                        subTopics[1] = "client/mqttx_d4c106a3";
                        break;
                    case "untiy1":
                        topic = "client/mqttx_d4c106a3";
                        subTopics[0] = "client/manage";
                        subTopics[1] = "client/mqttx_d4c106a3";
                        break;
                    case "untiy2":
                        topic = "client/mqttx_b4c02ddc";
                        subTopics[0] = "client/manage";
                        subTopics[1] = "client/mqttx_b4c02ddc";
                        break;
                    case "untiy3":
                        topic = "client/mqttx_ab869d21";
                        subTopics[0] = "client/manage";
                        subTopics[1] = "client/mqttx_ab869d21";
                        break;
                }
                UnsubscribeTopics();
                SubscribeTopics();
            }
        }
    }
}
