/*
The MIT License (MIT)

Copyright (c) 2018 Giovanni Paolo Vigano'

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
r
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using UnityEngine;
using uPLibrary.Networking.M2Mqtt;
using uPLibrary.Networking.M2Mqtt.Messages;

/// <summary>
/// Adaptation for Unity of the M2MQTT library (https://github.com/eclipse/paho.mqtt.m2mqtt),
/// modified to run on UWP (also tested on Microsoft HoloLens).
/// </summary>

namespace M2MqttUnity
{
    /// <summary>
    /// 普通MonoBehavior类包装MQTT客户端，使用双缓冲区来推迟主线程中的消息处理
    /// </summary>
    public class M2MqttUnityClient : MonoBehaviour
    {
        [Header("MQTT代理配置")]
        public string brokerAddress = ""; // IP地址或URL
        public int brokerPort;// 端口
        public bool isEncrypted = false; // 使用加密连接
        [Header("连接参数")]
        public int connectionDelay = 500; // 到代理的连接延迟了给定的毫秒
        public int timeoutOnConnection = MqttSettings.MQTT_CONNECT_TIMEOUT; // 连接超时（毫秒）
        public bool autoConnect = false; // 启动时自动连接
        public string mqttUserName = null; // MQTT代理的用户名。如果不需要用户名,请保留为空
        public string mqttPassword = null; // MQTT代理的密码。如果不需要密码,请保持空

        /// <summary>
        /// 包装MQTT客户端
        /// </summary>
        protected MqttClient client;

        private List<MqttMsgPublishEventArgs> messageQueue1 = new List<MqttMsgPublishEventArgs>();
        private List<MqttMsgPublishEventArgs> messageQueue2 = new List<MqttMsgPublishEventArgs>();
        private List<MqttMsgPublishEventArgs> frontMessageQueue = null; // 前端消息队列
        private List<MqttMsgPublishEventArgs> backMessageQueue = null; // 返回消息队列
        private bool mqttClientConnectionClosed = false;
        private bool mqttClientConnected = false;

        /// <summary>
        /// 成功建立连接时激发的事件
        /// </summary>
        public event Action ConnectionSucceeded;
        /// <summary>
        /// 连接失败时激发的事件
        /// </summary>
        public event Action ConnectionFailed;

        /// <summary>
        /// 使用当前设置连接到代理
        /// </summary>
        public virtual void Connect()
        {
            if (client == null || !client.IsConnected)
            {
                ThreadStart ts = new ThreadStart(DoConnect);
                Thread th = new Thread(ts);
                th.Start();
              //  StartCoroutine(DoConnect());
            }
        }

        /// <summary>
        /// 断开连接（如果已连接)
        /// </summary>
        public virtual void Disconnect()
        {
            if (client != null)
            {
                StartCoroutine(DoDisconnect());
            }
        }

        /// <summary>
        /// 覆盖此方法以在连接前采取一些操作（例如显示消息）
        /// </summary>
        protected virtual void OnConnecting()
        {
            Debug.LogFormat("正在连接 {0}:{1}...\n", brokerAddress, brokerPort.ToString());
        }

        /// <summary>
        /// 如果连接成功，请重写此方法以执行某些操作。
        /// </summary>
        protected virtual void OnConnected()
        {
            Debug.LogFormat("已经连接到 {0}:{1}...\n", brokerAddress, brokerPort.ToString());

            SubscribeTopics();

            if (ConnectionSucceeded != null)
            {
                ConnectionSucceeded();
            }
        }

        /// <summary>
        /// 如果连接失败，请重写此方法以执行某些操作
        /// </summary>
        protected virtual void OnConnectionFailed(string errorMessage)
        {
            Debug.LogError("MQTT连接失败");
            if (ConnectionFailed != null)
            {
                ConnectionFailed();
            }
        }

        /// <summary>
        /// 重写此方法以订阅MQTT主题
        /// </summary>
        protected virtual void SubscribeTopics()
        {
        }

        /// <summary>
        /// 重写此方法以取消订阅MQTT主题（它们应该与您使用SubscribeTopics（）订阅的主题相同）
        /// </summary>
        protected virtual void UnsubscribeTopics()
        {
        }

        /// <summary>
        /// 在应用程序退出之前断开连接
        /// </summary>
        protected virtual void OnApplicationQuit()
        {
            CloseConnection();
        }

        /// <summary>
        /// 初始化MQTT消息队列
        /// 如果重写此方法，请记住调用base.Awake（）
        /// </summary>
        protected virtual void Awake()
        {
            frontMessageQueue = messageQueue1;
            backMessageQueue = messageQueue2;
        }

        /// <summary>
        /// 如果autoConnect设置为true，则在启动时连接
        /// </summary>
        protected virtual void Start()
        {
            if (autoConnect)
            {
                Connect();
            }
        }

        /// <summary>
        /// 解码消息
        /// 为每个需要处理的接收消息覆盖此方法
        /// </summary>
        protected virtual void DecodeMessage(string topic, byte[] message)
        {
            Debug.LogFormat("Message received on topic: {0}", topic);
        }

        /// <summary>
        /// 断开连接
        /// 覆盖此方法以在断开连接时执行一些操作
        /// </summary>
        protected virtual void OnDisconnected()
        {
            Debug.LogError("Disconnected.");
        }

        /// <summary>
        /// 连接丢失时
        /// 覆盖此方法以在连接关闭时执行一些操作。
        /// </summary>
        protected virtual void OnConnectionLost()
        {
            Debug.LogError("CONNECTION LOST!");
        }

        /// <summary>
        /// 在主线程中，收入消息和事件的处理被推迟
        /// 如果重写了ProcessMqttEvents（）方法，请记住在Update（）方法中调用它
        /// </summary>
        protected virtual void Update()
        {
            ProcessMqttEvents();
        }

        /// <summary>
        /// 处理Mqtt事件
        /// </summary>
        protected virtual void ProcessMqttEvents()
        {
            // 处理主队列中的消息
            SwapMqttMessageQueues();
            ProcessMqttMessageBackgroundQueue();
            // 同时处理消息收入
            SwapMqttMessageQueues();
            ProcessMqttMessageBackgroundQueue();

            if (mqttClientConnectionClosed)
            {
                mqttClientConnectionClosed = false;
                OnConnectionLost();
            }
        }

        private void ProcessMqttMessageBackgroundQueue()
        {
            foreach (MqttMsgPublishEventArgs msg in backMessageQueue)
            {
                DecodeMessage(msg.Topic, msg.Message);
            }
            backMessageQueue.Clear();
        }

        /// <summary>
        /// 交换消息队列以在处理队列时继续接收消息。
        /// </summary>
        private void SwapMqttMessageQueues()
        {
            frontMessageQueue = frontMessageQueue == messageQueue1 ? messageQueue2 : messageQueue1;
            backMessageQueue = backMessageQueue == messageQueue1 ? messageQueue2 : messageQueue1;
        }

        private void OnMqttMessageReceived(object sender, MqttMsgPublishEventArgs msg)
        {
            frontMessageQueue.Add(msg);
        }

        private void OnMqttConnectionClosed(object sender, EventArgs e)
        {
            // 仅在连接时设置意外连接关闭（避免在受控断开的情况下进行事件处理）
            mqttClientConnectionClosed = mqttClientConnected;
            mqttClientConnected = false;
        }

        /// <summary>
        /// 使用当前设置连接到代理。
        /// </summary>
        /// <returns>执行是在协程中完成的</returns>
        private void DoConnect()
        {
            // 等待给定的延迟
            // yield return new WaitForSecondsRealtime(connectionDelay / 1000f);
            // 给Unity留点时间刷新UI
            // yield return new WaitForEndOfFrame();

            // create client instance
            if (client == null)
            {
#if (!UNITY_EDITOR && UNITY_WSA_10_0 && !ENABLE_IL2CPP)
                client = new MqttClient(brokerAddress,brokerPort,isEncrypted, isEncrypted ? MqttSslProtocols.SSLv3 : MqttSslProtocols.None);
#else
                string folderPath;
                if(File.Exists(Application.streamingAssetsPath + "/emqxsl-ca.crt"))
                    folderPath = Application.streamingAssetsPath + "/emqxsl-ca.crt";
                else
                    folderPath = Path.Combine("Assets", "Samples", "MQTT", "1.0.0", "GHZMQTT", "StreamingAssets","emqxsl-ca.crt");
                X509Certificate cert = new X509Certificate(folderPath);
                Debug.Log("IP地址或URL==>"+ brokerAddress);
                client = new MqttClient(brokerAddress, brokerPort, isEncrypted, cert, cert, isEncrypted ? MqttSslProtocols.SSLv3 : MqttSslProtocols.None);
                //System.Security.Cryptography.X509Certificates.X509Certificate cert = new System.Security.Cryptography.X509Certificates.X509Certificate();
                // client = new MqttClient(brokerAddress, brokerPort, isEncrypted, cert,userCertificateSelectionCallback, MyRemoteCertificateValidationCallback);
#endif
            }
            else if (client.IsConnected)
            {
                Debug.LogError("IsConnected");
             //   yield break;
            }
            OnConnecting();

            // 给Unity留点时间刷新UI
            // yield return new WaitForEndOfFrame();
            // yield return new WaitForEndOfFrame();

            client.Settings.TimeoutOnConnection = timeoutOnConnection;
            string clientId = Guid.NewGuid().ToString();
            Debug.LogError("自动生成clientId:"+clientId);
            clientId = "mqttx_ab869d21";
            client.Connect(clientId, mqttUserName, mqttPassword);

            if (client.IsConnected)
            {
                client.ConnectionClosed += OnMqttConnectionClosed;
                // register to message received
                client.MqttMsgPublishReceived += OnMqttMessageReceived;
                mqttClientConnected = true;
                OnConnected();
            }
            else
            {
                OnConnectionFailed("CONNECTION FAILED!");
            }
        }
        /// <summary>
        /// 用户证书选择回调
        /// </summary>
        private bool userCertificateSelectionCallback(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            throw new NotImplementedException();
        }
        /// <summary>
        /// 我的远程证书验证回调
        /// </summary>
        private bool MyRemoteCertificateValidationCallback(object sender, X509Certificate certificate, X509Chain chain, SslPolicyErrors sslPolicyErrors)
        {
            throw new NotImplementedException();
        }

        private IEnumerator DoDisconnect()
        {
            yield return new WaitForEndOfFrame();
            CloseConnection();
            OnDisconnected();
        }

        private void CloseConnection()
        {
            mqttClientConnected = false;
            if (client != null)
            {
                if (client.IsConnected)
                {
                    UnsubscribeTopics();
                    client.Disconnect();
                }
                client.MqttMsgPublishReceived -= OnMqttMessageReceived;
                client.ConnectionClosed -= OnMqttConnectionClosed;
                client = null;
            }
        }

#if ((!UNITY_EDITOR && UNITY_WSA_10_0))
        private void OnApplicationFocus(bool focus)
        {
            // 在UWP10（HoloLens）上，我们无法判断应用程序是真的关闭了还是最小化了
            // (https://forum.unity.com/threads/onapplicationquit-and-ondestroy-are-not-called-on-uwp-10.462597/)
            if (focus)
            {
                Connect();
            }
            else
            {
                CloseConnection();
            }
        }
#endif
    }
}
